(* file: numerals.ml
   author: Bob Muller

   CS1103 Computer Science I Honors

   Lecture Notes/Code for Week 8

  1. Review of recursion
  2. Numeral Systems

  *******************************************************************

  1. Review of recursion.

  A recursive solution is closely related to an inductive proof in
  mathematics. Imagine crossing a stream on stepping stones. Recursion
  allows you to assume that you are standing on the last stone --- all
  that you have to do is get from that stone to the next spot.

  Two examples:

   powerList : 'a list -> ('a list) list

   The call (powerList xs) produces all sublists of xs.
*)
let rec powerList xs =
  match xs with
  | [] -> [[]]
  | x :: xs ->
     let almost = powerList xs   (* We assume a solution to the slightly *)
     in                          (* smaller problem. *)
     almost @ (List.map (fun ys -> x :: ys) almost)

(* permutations : int -> 'a list -> ('a list) list

   The call (permutations n xs) returns a list of all n-length
  permutations of the symbols in xs.
*)
let rec permutations n symbols =
  match n = 0 with
  | true  -> [[]]
  | false ->
    let almost = permutations (n - 1) symbols
    in
    List.fold_left
      (@)
      []
      (List.map (fun sym -> List.map (fun alm -> sym :: alm) almost) symbols)


(* Addition in positional numeral systems.

  digitValue : int -> int -> int -> int
*)
let digitValue digit base n =
  digit * int_of_float ((float base) ** (float n))

(* toDecminal : int list -> int -> int
*)
let rec toDecimal digits base =
  match digits with
  | [] -> 0
  | digit :: digits ->
    let digitVal = digitValue digit base (List.length digits)
    in
    digitVal + (toDecimal digits base)

let div m n = (m / n, m mod n)

(* decimalTo : int -> int -> int list
*)
let decimalTo numeral base =
  let rec repeat numeral digits =
    match numeral = 0 with
    | true  -> digits
    | false ->
      let (quotient, remainder) = div numeral base
      in
      repeat quotient (remainder :: digits)
  in
  repeat numeral []

(* NB: (decimalTo (toDecimal digits base) base) = digits *)

(* positionalAdd : int list -> int list -> int -> int list
*)
let positionalAdd ms ns base =
  let rec repeat ms ns carryIn answer =
    match (ms, ns) with
    | ([], []) -> answer
    | (m :: ms, n :: ns) ->
      let sum = m + n + carryIn in
      let (carryOut, sum) = div sum base
      in
      repeat ms ns carryOut (sum :: answer)
    | _ -> failwith "cannot happen"
  in
  repeat (List.rev ms) (List.rev ns) 0 []
